"""Denoising utilities."""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec

from .matrix import _check_weights, fold, theshapeof, unfold


def demean(X, weights=None, return_mean=False, inplace=False):
    """Remove weighted mean over rows (samples).

    Parameters
    ----------
    X : array, shape=(n_samples, n_channels[, n_trials])
        Data.
    weights : array, shape=(n_samples)
    return_mean : bool
        If True, also return signal mean (default=False).
    inplace : bool
        If True, save the resulting array in X (default=False).

    Returns
    -------
    X : array, shape=(n_samples, n_channels[, n_trials])
        Centered data.
    mn : array
        Mean value.

    """
    weights = _check_weights(weights, X)
    ndims = X.ndim

    if not inplace:
        X = X.copy()

    n_samples, n_chans, n_trials = theshapeof(X)
    X = unfold(X)

    if weights.any():
        weights = unfold(weights)

        if weights.shape[0] != X.shape[0]:
            raise ValueError("X and weights arrays should have same " +
                             "number of samples (rows).")

        if weights.shape[1] == 1 or weights.shape[1] == n_chans:
            mn = (np.sum(X * weights, axis=0) /
                  np.sum(weights, axis=0))[None, :]
        else:
            raise ValueError("Weight array should have either the same " +
                             "number of columns as X array, or 1 column.")

    else:
        mn = np.mean(X, axis=0, keepdims=True)

    # Remove mean (no copy)
    X -= mn

    if n_trials > 1 or ndims == 3:
        X = fold(X, n_samples)

    if return_mean:
        return X, mn  # the_mean.shape=(1, the_mean.shape[0])
    else:
        return X


def mean_over_trials(X, weights=None):
    """Compute weighted mean over trials (last dimension)."""
    if weights is None:
        weights = np.array([])
    else:
        weights = _check_weights(weights, X)
    n_samples, n_chans, n_trials = theshapeof(X)

    if not weights.any():
        y = np.mean(X, 2)
        tw = np.ones((n_samples, n_chans, 1))
    else:
        m, n, o = theshapeof(weights)
        if m != n_samples:
            raise RuntimeError("weights matrix has incorrect n_samples")
        if o != n_trials:
            raise RuntimeError("weights matrix has incorrect n_trials")

        # Unfold to (n_times*n_trials, n_chans), multiply by weights and refold
        X = unfold(X)
        weights = unfold(weights)
        X = X * weights
        X = fold(X, n_samples)
        weights = fold(weights, n_samples)

        # Take weighted average
        with np.errstate(divide="ignore", invalid="ignore"):
            y = np.sum(X, -1) / np.sum(weights, -1)
        tw = np.mean(weights, -1)
        y = np.nan_to_num(y)
    return y, tw


def wpwr(X, weights=None):
    """Weighted power."""
    if weights is None:
        weights = np.array([])

    X = unfold(X)
    weights = unfold(weights)

    if weights.size > 0:
        X = X * weights
        y = np.sum(X ** 2)
        tweight = np.sum(weights)
    else:
        y = np.sum(X ** 2)
        tweight = X.size

    return y, tweight


def find_outlier_samples(X, toobig1, toobig2=[]):
    """Find outlier trials using an absolute threshold."""
    n_samples, n_chans, n_trials = theshapeof(X)
    X = unfold(X)

    # apply absolute threshold
    weights = np.ones((n_trials * n_samples, n_chans))

    if toobig1 is not None:
        weights[np.where(abs(X) > toobig1)] = 0
        X = demean(X, weights)

        weights[np.where(abs(X) > toobig1)] = 0
        X = demean(X, weights)

        weights[np.where(abs(X) > toobig1)] = 0
        X = demean(X, weights)
    else:
        weights = np.ones(X.shape)

    # apply relative threshold
    if toobig2:
        _, mn = demean(X ** 2, weights, return_mean=True)
        X = np.tile(mn, (X.shape[0], 1))
        idx = np.where(X ** 2 > (X * toobig2))[0]
        weights[idx] = 0

    weights = fold(weights, n_samples)

    return weights


def find_outlier_trials(X, thresh=None, show=True):
    """Find outlier trials.

    For example thresh=2 rejects trials that deviate from the mean by
    more than twice the average deviation from the mean.

    Parameters
    ----------
    X : ndarray, shape=(n_times, n_chans[, n_trials])
        Data array.
    thresh : float or array of floats
        Keep trials less than thresh from mean.
    show : bool
        If True (default), plot trial deviations before and after.

    Returns
    -------
    bads : list of int
        Indices of trials to reject.
    d : array
        Relative deviations from mean.

    """
    if thresh is None:
        thresh = np.array([np.inf])
    elif isinstance(thresh, (float, int)):
        thresh = np.array([thresh])
    else:
        thresh = np.asarray(thresh)

    if X.ndim > 3:
        raise ValueError("X should be 2D or 3D")
    elif X.ndim == 3:
        n_samples, n_chans, n_trials = theshapeof(X)
        X = np.reshape(X, (n_samples * n_chans, n_trials))
    else:
        n_chans, n_trials = X.shape

    avg = np.nanmean(X, axis=-1, keepdims=True)  # mean over trials
    d = X - avg  # difference from mean
    d = np.nansum(d ** 2, axis=0)

    d = d / (np.nansum(X ** 2) / n_trials)
    idx = np.where(d < thresh[0])[0]

    if show:
        plt.figure(figsize=(7, 4))
        gs = gridspec.GridSpec(1, 2)
        plt.suptitle("Outlier trial detection")

        # Before
        ax1 = plt.subplot(gs[0, 0])
        ax1.plot(d, ls="-")
        ax1.plot(np.setdiff1d(np.arange(n_trials), idx),
                 d[np.setdiff1d(np.arange(n_trials), idx)], color="r", ls=" ",
                 marker=".")
        ax1.axhline(y=thresh[0], color="grey", linestyle=":")
        ax1.set_xlabel("Trial #")
        ax1.set_ylabel("Normalized deviation from mean")
        ax1.set_title("Before, " + str(len(d)))
        ax1.set_xlim(0, len(d) + 1)
        plt.draw()

        # After
        ax2 = plt.subplot(gs[0, 1])
        _, dd = find_outlier_trials(X[:, idx], None, False)
        ax2.plot(dd, ls="-")
        ax2.set_xlabel("Trial #")
        ax2.set_title("After, " + str(len(idx)))
        ax2.yaxis.tick_right()
        ax2.set_xlim(0, len(idx) + 1)
        plt.show()

    thresh = thresh[1:]
    if thresh.any():
        bads, _ = find_outlier_trials(X[:, idx], thresh, show)
        idx = np.setdiff1d(idx, idx[bads])

    bads_accumulated = []
    if len(idx) < n_trials:
        bads_accumulated = np.setdiff1d(range(n_trials), idx)

    return bads_accumulated, d



"""Statistics utilities."""

import numpy as np

from .matrix import theshapeof

try:
    import mne
except ImportError:
    mne = None

rng = np.random.default_rng()


def rms(X, axis=0):
    """Root-mean-square along given axis."""
    return np.sqrt(np.mean(X ** 2, axis=axis, keepdims=True))


def robust_mean(X, axis=0, percentile=[5, 95]):
    """Do robust mean based on JR Kings implementation."""
    X = np.array(X)
    axis_ = axis
    # force axis to be 0 for facilitation
    if axis is not None and axis != 0:
        X = np.transpose(X, [axis] + range(0, axis) + range(axis + 1, X.ndim))
        axis_ = 0
    mM = np.percentile(X, percentile, axis=axis_)
    indices_min = np.where((X - mM[0][np.newaxis, ...]) < 0)
    indices_max = np.where((X - mM[1][np.newaxis, ...]) > 0)
    X[indices_min] = np.nan
    X[indices_max] = np.nan
    m = np.nanmean(X, axis=axis_)
    return m


def rolling_corr(X, y, window=None, sfreq=1, step=1, axis=0):
    """Calculate rolling correlation between some data and a reference signal.

    Parameters
    ----------
    X: array, shape=(n_times, n_chans[, n_epochs])
        Test signal. First dimension should be time.
    y: array, shape=(n_times[, n_epochs])
        Reference signal.
    window : int
        Number of timepoints for to include for each correlation calculation.
    sfreq: int
        Sampling frequency (default=1).
    step : int
        If > 1, only compute correlations every `step` samples.

    Returns
    -------
    corr: array, shape=(n_times - window, n_channels[, n_epochs])
        Rolling window correlation.
    t_corr : array
        Corresponding time vector.

    """
    if window is None:
        window = X.shape[0] - 1
    if y.ndim == 3:
        y = np.squeeze(y)
    if X.ndim > 3:
        raise AttributeError("Data must be 2D or 3D.")
    if y.shape[0] != X.shape[0]:
        raise AttributeError("X and y must share the same time axis.")
    if y.ndim > 2:
        raise AttributeError("y must be at most 2D.")

    n_times, n_chans, n_epochs = theshapeof(X)
    timebins = np.arange(n_times - window, 0, -step)[::-1]

    corr = np.zeros((len(timebins), n_chans, n_epochs))
    for i, t in enumerate(timebins):
        for ep in range(n_epochs):
            epy = np.take(y[t:t + window, ...], ep, -1).squeeze()
            for ch in range(n_chans):
                epx = np.take(X[t:t + window, ch, ...], ep, -1).squeeze()
                corr[i, ch, ep] = np.corrcoef(epx, epy)[0, 1]

    if n_epochs == 1:
        corr = corr.squeeze(-1)

    # Times relative to end of window
    t_corr = (timebins + window) / float(sfreq)

    assert len(t_corr) == corr.shape[0]

    return corr, t_corr


def bootstrap_ci(X, n_bootstrap=2000, ci=(5, 95), axis=-1):
    """Confidence intervals from non-parametric bootstrap resampling.

    Bootstrap is computed over the chosen axis, *with* replacement. By default,
    the resampling is performed over the last dimension (trials), but this can
    also be done over channels (axis=1).

    Parameters
    ----------
    X : array, shape=(n_times, n_chans[, n_trials])
        Input data.
    n_bootstrap : int
        Number of bootstrap resamplings (default=2000). For publication
        quality results, it is usually recommended to have more than 5000
        iterations.
    ci : length-2 tuple
        Confidence interval (default=(5, 95)).
    axis : int
        Axis to operate on.

    Returns
    -------
    ci_low, ci_up : arrays, shape=(n_times, n_chans)
        Confidence intervals.

    """
    idx = np.arange(X.shape[axis], dtype=int)
    shape = list(X.shape)
    shape.pop(axis)

    bootstraps = np.nan * np.ones((n_bootstrap,) + tuple(shape))
    for i in range(n_bootstrap):
        temp_idx = rng.choice(idx, replace=True, size=len(idx))
        bootstraps[i] = np.mean(np.take(X, temp_idx, axis=axis), axis=axis)

    ci_low, ci_up = np.percentile(bootstraps, ci, axis=0)

    return ci_low, ci_up


def bootstrap_snr(epochs, n_bootstrap=2000, baseline=None, window=None):
    """Get bootstrap SNR from non-parametric bootstrap resampling.

    The lower bound SNR value can serve as a minimum threshold of signal
    quality. Parks et al. (2016) suggest using a threshold of 3 dB when
    assessing whether to remove subjects from ERP studies.

    Parameters
    ----------
    epochs : mne.Epochs instance
        Epochs instance to compute SNR from.
    n_bootstrap : int
        Number of bootstrap iterations (should be > 10000 for publication
        quality).
    baseline : tuple or list of length 2, or None
        The time interval to apply rescaling / baseline correction. If None do
        not apply it. If baseline is ``(bmin, bmax)`` the interval is between
        ``bmin`` (s) and ``bmax`` (s). If ``bmin is None`` the beginning of the
        data is used and if ``bmax is None`` then ``bmax`` is set to the end of
        the interval. If baseline is ``(None, None)`` the entire time interval
        is used. If baseline is None, no correction is applied.
    window : tuple or list of length 2, or None
        The time interval used to compute the numerator of the SNR. If window
        is ``(bmin, bmax)`` the interval is between ``bmin`` (s) and ``bmax``
        (s). If ``bmin is None`` the beginning of the data is used and if
        ``bmax is None`` then ``bmax`` is set to the end of the interval. If
        window is ``(None, None)`` the entire time interval is used. If None
        use all positive times.

    Returns
    -------
    ERP : length-3 tuple
        Mean bootstrap ERP, with 90% confidence intervals.
    SNR : length-3 tuple
        Mean bootstrap SNR, with 90% confidence intervals.

    References
    ----------
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4751267/

    """
    indices = np.arange(len(epochs.selection), dtype=int)
    n_chans = len(epochs.ch_names)
    erp_bs = np.empty((n_bootstrap, n_chans, len(epochs.times)))
    gfp_bs = np.empty((n_bootstrap, n_chans, len(epochs.times)))

    for i in range(n_bootstrap):
        bs_indices = rng.choice(indices, replace=True, size=len(indices))
        erp_bs[i] = np.mean(epochs._data[bs_indices, ...], 0)

        # Baseline correct mean waveform
        if baseline:
            erp_bs[i] = mne.baseline.rescale(erp_bs[i], epochs.times,
                                             baseline=baseline,
                                             verbose="ERROR")

        # Rectify waveform
        gfp_bs[i] = np.sqrt(erp_bs[i] ** 2)

    # Bootstrap ERP confidence intervals
    ci_low, ci_up = np.percentile(erp_bs, (10, 90), axis=0)

    # Calculate SNR for each bootstrapped ERP; form distribution in `snr_dist`
    snr_dist = np.zeros((n_bootstrap, n_chans))
    if window is not None:
        if window[0] is None:
            window[0] = 0
        if window[1] is None:
            window[1] = np.inf
        post = np.logical_and(epochs.times >= window[0],
                              epochs.times <= window[1])
    else:
        post = epochs.times >= 0
    pre = epochs.times <= 0

    # SNR for each bootstrap iteration
    snr_dist = 20 * np.log10(gfp_bs[..., post].mean(-1) /
                             gfp_bs[..., pre].mean(-1))

    # Mean, lower, and upper bound SNR
    snr_low, snr_up = np.percentile(snr_dist, (10, 90), axis=0)
    snr_mean = np.mean(snr_dist, axis=0)
    mean_bs_erp = np.mean(erp_bs, axis=0)

    return (mean_bs_erp, ci_low, ci_up), (snr_mean, snr_low, snr_up)


def cronbach(epochs, K=None, n_bootstrap=2000, tmin=None, tmax=None):
    r"""Compute reliability of ERP.

    Internal reliability of the ERN and Pe as a function of increasing trial
    numbers can be quantified with Cronbach's alpha:

        $$\alpha = \frac{K}{K-1} \left(1-\frac{\sum_{i=1}^K
          \sigma^2_{Y_i}}{ \sigma^2_X}\right)$$

    Hinton, Brownlow, McMurray, and Cozens (2004) have suggested that
    Cronbach's alpha exceeding .90 indicates excellent internal reliability,
    between .70 and .90 indicates high internal reliability, from .50 to .70
    indicates moderate internal reliability, and below .50 is low.

    Parameters
    ----------
    epochs : mne.Epochs | ndarray, shape=(n_trials, n_chans, n_samples)
        Epochs to compute alpha from.
    K : int
        Number of trials to use for alpha computation.
    n_bootstrap : int
        Number of bootstrap resamplings.
    tmin : float
        Start time of epoch.
    tmax : float
        End of epoch.

    Returns
    -------
    alpha : array, shape=(n_chans,)
        Cronbach alpha value
    bounds: length-2 tuple
        Lower and higher bound of CI.

    """
    if isinstance(epochs, np.ndarray):
        erp = epochs
        tmin = tmin if tmin else 0
        tmax = tmax if tmax else -1
    elif isinstance(epochs, mne.BaseEpochs):
        erp = epochs.get_data()
        tmin = epochs.time_as_index(tmin)[0] if tmin else 0
        tmax = epochs.time_as_index(tmax)[0] if tmax else -1
    else:
        raise ValueError("epochs must be an mne.Epochs or numpy array.")

    n_trials, n_chans, n_samples = erp.shape
    if K is None:
        K = n_trials

    alpha = np.empty((n_bootstrap, n_chans))
    for b in np.arange(n_bootstrap):  # take K trials randomly
        idx = rng.choice(range(n_trials), K)
        X = erp[idx, :, tmin:tmax]
        sigmaY = X.var(axis=2).sum(0)  # var over time
        sigmaX = X.sum(axis=0).var(-1)  # var of average
        alpha[b] = K / (K - 1) * (1 - sigmaY / sigmaX)

    ci_lo, ci_hi = np.percentile(alpha, (10, 90), axis=0)
    return alpha.mean(0), ci_lo, ci_hi


def snr_spectrum(X, freqs, n_avg=1, n_harm=1, skipbins=1):
    """Compute Signal-to-Noise-corrected spectrum.

    The implementation tries to replicate examples in [1; 2; 3]_.

    Parameters
    ----------
    X : ndarray , shape=(n_freqs, n_chans,[ n_trials,])
        One-sided power spectral density estimate, specified as a real-valued,
        nonnegative array. The power spectral density must be expressed in
        linear units, not decibels.
    freqs : array, shape=(n_freqs,)
        Frequency bins.
    n_avg : int
        Number of neighbour bins to estimate noise over. Make sure that this
        value doesn't overlap with neighbouring target frequencies.
    n_harm : int
        Compute SNR at each frequency bin as a pooled RMS over this bin and
        n_harm harmonics (see references below).
    skipbins : int
        Number of bins skipped to estimate noise of neighbouring bins.

    Returns
    -------
        SNR : ndarray, shape=(n_freqs, n_chans, n_trials) or (n_freqs, n_chans)
            Signal-to-Noise-corrected spectrum.

    References
    ----------
    .. [1] Cottereau, B. R., McKee, S. P., Ales, J. M., & Norcia, A. M. (2011).
       Disparity-tuned population responses from human visual cortex. The
       Journal of Neuroscience, 31(3), 954-965.
    .. [2] Cottereau, B. R., McKee, S. P., & Norcia, A. M. (2014). Dynamics and
       cortical distribution of neural responses to 2D and 3D motion in human.
       Journal of neurophysiology, 111(3), 533-543.
    .. [3] de Heering, A., & Rossion, B. (2015). Rapid categorization of
       natural face images in the infant right hemisphere. Elife, 4.

    """
    if X.ndim == 3:
        n_freqs = X.shape[0]
        n_chans = X.shape[1]
        n_trials = X.shape[-1]
    elif X.ndim == 2:
        n_trials = 1
        n_freqs = X.shape[0]
        n_chans = X.shape[1]
    else:
        raise ValueError("Data must have shape (n_freqs, n_chans, [n_trials,])"
                         f", got {X.shape}")

    # Number of points to get desired resolution
    X = np.reshape(X, (n_freqs, n_chans * n_trials))
    SNR = np.zeros_like(X)

    for i_bin in range(n_freqs):

        # Indices of bins to average over
        # ----------------------------------------------------------------------
        # bin_peaks = np.zeros(n_harm, dtype=int)
        bin_noise = []
        bin_peaks = []
        for h in range(n_harm):  # loop over harmonics

            # make sure freq <= fmax
            if freqs[i_bin] > 0 and freqs[i_bin] * (h + 1) <= freqs[-1]:

                # Get indices of harmonics
                bin_peaks.append(
                    int(np.argmin(np.abs(freqs[i_bin] * (h + 1) - freqs)))
                )

                # Now get indices of noise (i.e., neighbouring FFT bins)
                # eg if currentbin=54, navg=3, skipbins=1 :
                # bin_noise = 51, 52, 56, 57
                tmp = np.r_[
                    (np.arange(bin_peaks[h] - skipbins - n_avg,
                               bin_peaks[h] - skipbins),
                     np.arange(bin_peaks[h] + skipbins + 1,
                               bin_peaks[h] + skipbins + 1 + n_avg))]
                tmp = tmp.astype(int)

                # Remove impossible bin values (eg <1 or >n_samp)
                tmp = [t for t in tmp if t >= 0 and t < n_freqs]
                bin_noise.append(tmp)
                del tmp
            else:
                bin_peaks.append(0)
                bin_noise.append(0)

        # SNR at central bin is ratio between (power at central
        # bin) to (average of N surrounding bins)
        # --------------------------------------------------------------------------
        for i_trial in range(n_chans * n_trials):

            # Mean of signal over fundamental+harmonics
            A = np.mean(X[bin_peaks, i_trial] ** 2)

            # Noise around fundamental+harmonics
            B = np.zeros(len(bin_noise))
            for h in range(len(B)):
                B[h] = np.mean(X[bin_noise[h], i_trial].flatten() ** 2)

            # Ratio
            with np.errstate(divide="ignore", invalid="ignore"):
                SNR[i_bin, i_trial] = np.sqrt(A) / np.sqrt(B)

            del A
            del B

    # SNR[np.abs(SNR) == np.inf] = 1
    # SNR[SNR == 0] = 1
    # SNR[np.isnan(SNR)] = 1

    # Reshape matrix if necessary
    if np.min((n_trials, n_chans)) > 1:
        SNR = np.reshape(SNR, (n_freqs, n_chans, n_trials))

    return SNR



"""Test signal utils."""
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import butter, freqz, hilbert, lfilter

from meegkit.phase import ECHT
from meegkit.utils.sig import stmcb, teager_kaiser

rng = np.random.default_rng(9)

def test_teager_kaiser(show=False):
    """Test Teager-Kaiser Energy."""
    x = 2 * rng.random((1000, 2)) - 1
    x[100, 0] = 5
    x[200, 0] = 5
    x += np.cumsum(rng.standard_normal((1000, 1)) + 0.1, axis=0) / 1000
    for i in range(1, 5):
        print(i)
        y = teager_kaiser(x, M=i, m=1)
        assert y.shape[0] == 1000 - 2 * i

    if show:
        import matplotlib.pyplot as plt
        plt.figure()
        # plt.plot((x[1:, 0] - x[1:, 0].mean()) / np.nanstd(x[1:, 0]))
        # plt.plot((y[..., 0] - y[..., 0].mean()) / np.nanstd(y[..., 0]))
        plt.plot(x[1:, 0], label="X")
        plt.plot(y[..., 0], label="Y")
        plt.legend()
        plt.show()


def test_stcmb(show=True):
    """Test stcmb."""
    x = np.arange(100) < 1
    [b, a] = butter(6, 0.2)     # Butterworth filter design
    y = lfilter(b, a, x)        # Filter data using above filter

    w, h = freqz(b, a, 128)     # Frequency response
    [bb, aa] = stmcb(y, u_in=None, q=4, p=4, niter=5)
    ww, hh = freqz(bb, aa, 128)

    if show:
        import matplotlib.pyplot as plt
        f, ax = plt.subplots(2, 1)
        ax[0].plot(x, label="step")
        ax[0].plot(y, label="filt")
        ax[1].plot(w, np.abs(h), label="real")
        ax[1].plot(ww, np.abs(hh), label="stcmb")
        ax[0].legend()
        ax[1].legend()
        plt.show()

    np.testing.assert_allclose(h, hh, rtol=2)  # equal to 2%


def test_echt(show=False):
    """Test Endpoint-corrected Hilbert transform (ECHT) phase estimation."""
    rng = np.random.default_rng(38872)

    # Build data
    # -------------------------------------------------------------------------
    # First, we generate a multi-component signal with amplitude and phase
    # modulations, as described in the paper [1]_.
    f0 = 2
    filt_BW = f0 / 2
    N = 500
    sfreq = 200
    time = np.linspace(0, N / sfreq, N)
    X = np.cos(2 * np.pi * f0 * time - np.pi / 4)
    phase_true = np.angle(hilbert(X))
    X += rng.normal(0, 0.5, N)  # Add noise

    # Compute phase and amplitude
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # We compute the Hilbert phase, as well as the phase obtained with the ECHT
    # filter.
    # phase_hilbert = np.angle(hilbert(X))  # Hilbert phase

    # Compute ECHT-filtered signal
    l_freq = f0 - filt_BW / 2
    h_freq = f0 + filt_BW / 2
    echt = ECHT(l_freq, h_freq, sfreq)

    Xf = echt.fit_transform(X)
    phase_echt = np.angle(Xf).squeeze()
    # phase_true = np.roll(phase_true, 1)
    if show:
        fig, ax = plt.subplots(3, 1, figsize=(8, 5))
        ax[0].plot(time, X)
        ax[0].set_xlabel("Time (s)")
        ax[0].set_title("Test signal")
        ax[0].set_ylabel("Amplitude")

        ax[1].plot(time, phase_true, label="True phase", ls=":")
        ax[1].plot(time, phase_echt, label="ECHT phase", lw=.5, alpha=.8)
        ax[1].set_title("Phase")
        ax[1].set_ylabel("Amplitude")
        ax[1].set_xlabel("Time (s)")
        ax[1].legend(loc="upper right", fontsize="small")

        ax[2].plot(time, np.unwrap(phase_true - phase_echt.squeeze()),
                   label="Phase error")
        ax[2].set_title("Phase error")
        ax[2].set_ylabel("Error")

        plt.tight_layout()
        plt.show()

    mae =  (np.abs(np.unwrap(phase_true - phase_echt.squeeze())) > np.pi / 6).sum() / N
    assert mae < 0.1, mae

def test_echt_nd():
    """Test ECHT with ndim > 1."""
    rng = np.random.default_rng(38872)

    # Build data
    # -------------------------------------------------------------------------
    # First, we generate a multi-component signal with amplitude and phase
    # modulations, as described in the paper [1]_.
    f0 = 2
    filt_BW = f0 / 2
    N = 500
    sfreq = 200
    time = np.linspace(0, N / sfreq, N)
    X = np.cos(2 * np.pi * f0 * time - np.pi / 4)[:, None]
    X = X + rng.normal(0, 0.5, (N, 2))  # Add noise

    l_freq = f0 - filt_BW / 2
    h_freq = f0 + filt_BW / 2
    echt = ECHT(l_freq, h_freq, sfreq)

    Xf = echt.fit_transform(X)
    phase_echt = np.angle(Xf)

    assert phase_echt.shape == (N, 2)


if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_teager_kaiser()
    # test_stcmb()
    # test_echt()


"""
DSS example
===========

Find the linear combinations of multichannel data that maximize repeatability
over trials.

Uses meegkit.dss0().
"""
import matplotlib.pyplot as plt
import numpy as np

from meegkit import dss
from meegkit.utils import fold, rms, tscov, unfold

rng = np.random.default_rng(5)

###############################################################################
# Create simulated data
# -----------------------------------------------------------------------------

# Data are time * channel * trials.
n_samples = 100 * 3
n_chans = 30
n_trials = 100
noise_dim = 20  # dimensionality of noise

# Source signal
source = np.hstack((
    np.zeros((n_samples // 3,)),
    np.sin(2 * np.pi * np.arange(n_samples // 3) / (n_samples / 3)).T,
    np.zeros((n_samples // 3,))))[np.newaxis].T
s = source * rng.standard_normal((1, n_chans))  # 300 * 30
s = s[:, :, np.newaxis]
s = np.tile(s, (1, 1, 100))

# Noise
noise = np.dot(
    unfold(rng.standard_normal((n_samples, noise_dim, n_trials))),
    rng.standard_normal((noise_dim, n_chans)))
noise = fold(noise, n_samples)

# Mix signal and noise
SNR = 0.1
data = noise / rms(noise.flatten()) + SNR * s / rms(s.flatten())

###############################################################################
# Apply DSS to clean them
# -----------------------------------------------------------------------------

# Compute original and biased covariance matrices
c0, _ = tscov(data)

# In this case the biased covariance is simply the covariance of the mean over
# trials
c1, _ = tscov(np.mean(data, 2))

# Apply DSS
[todss, _, pwr0, pwr1] = dss.dss0(c0, c1)
z = fold(np.dot(unfold(data), todss), epoch_size=n_samples)

# Find best components
best_comp = np.mean(z[:, 0, :], -1)

###############################################################################
# Plot results
# -----------------------------------------------------------------------------
f, (ax1, ax2, ax3) = plt.subplots(3, 1)
ax1.plot(source, label="source")
ax2.plot(np.mean(data, 2), label="data")
ax3.plot(best_comp, label="recovered")
plt.legend()
plt.show()



"""Sparse time-artefact removal."""
import numpy as np
from scipy.signal import filtfilt

from .utils import demean, fold, mrdivide, normcol, pca, theshapeof, tscov, unfold, wpwr


def star(X, thresh=1, closest=[], depth=1, pca_thresh=1e-15, n_smooth=10,
         min_prop=0.5, n_iter=3, verbose=True):
    """Sparse time-artifact removal.

    Parameters
    ----------
    X : array, shape=(n_times, n_chans[, n_trials])
        Data to denoise.
    thresh : float
        Threshold for eccentricity measure (default=1).
    closest : array, shape=(n_chans, n_neighbours)
        Indices of channels that are closest to each channel (default=all).
    depth : int
        Maximum number of channels to fix at each sample (default=1).
    pca_thresh : float
        Threshold for discarding weak PCs (default=1e-15)
    n_smooth : int
        Samples, smoothing to apply to eccentricity (default=10).
    min_prop : float
        Minimum proportion of artifact free at first iteration (default=0.3).
    n_iter : int
        Iterations to refine c0 (default=3).
    verbose : bool | 'debug'
        Verbosity. If 'debug', also draw some diagnostics plots.

    Returns
    -------
    y : array
        Denoised data.
    w : array, shape=(n_times, 1)
        0 for parts that needed fixing, 1 elsewhere.
    ww : array, shape=(n_times, n_chans)
        0 for parts that needed fixing, 1 elsewhere.

    See Also
    --------
    sns

    """
    if thresh is None:
        thresh = 1
    if len(closest) > 0 and closest.shape[0] != X.shape[1]:
        raise ValueError("`closest` should have as many rows as n_chans")

    rng = np.random.default_rng()

    ndims = X.ndim
    n_samples, n_chans, n_trials = theshapeof(X)
    X = unfold(X)

    p0, _ = wpwr(X)
    X, intercept = demean(X, return_mean=True)
    norm = np.sqrt(np.mean(X ** 2, axis=0))
    X = normcol(X)
    p00, _ = wpwr(X)

    # NaN and zero channels are set to rand, which effectively excludes them
    idx_nan = np.all(np.isnan(X), axis=0)
    idx_zero = np.all(X == 0, axis=0)
    if idx_nan.any():
        X[:, idx_nan] = rng.standard_normal((X.shape[0], np.sum(idx_nan)))
    if idx_zero.any():
        X[:, idx_zero] = rng.standard_normal((X.shape[0], np.sum(idx_zero)))

    # initial covariance estimate
    X = demean(X)
    c0, _ = tscov(X)

    # Phase 1
    # -------------------------------------------------------------------------
    # Find time intervals where at least one channel is eccentric -> w == 0
    # Compute covariance on artifact-free data.

    iter = n_iter
    while iter > 0:
        w = np.ones((X.shape[0],))
        d = np.zeros_like(X)
        for ch in np.arange(n_chans):
            neighbours = _closest_neighbours(closest, ch, n_chans)

            # Compute channel data estimated from its neighbours
            z = _project_channel(X[:, neighbours], c0, ch, neighbours)

            # Compute eccentricity over time
            d[:, ch] = _eccentricity(X[:, ch][:, None], z, w, n_smooth).T
            d[:, ch] = d[:, ch] / thresh

            # Aggregate weights over channels
            # w == 0 : artifactual sample
            # w == 1 : clean time sample
            w = np.min((w, (d[:, ch] < 1)), axis=0)  # w==0 for artifact part

        artifact_free = np.mean(w, axis=0)
        if verbose:
            print(f"proportion artifact free: {artifact_free:.2f}")

        if iter == n_iter and artifact_free < min_prop:
            thresh = thresh * 1.1
            if verbose:
                print(f"Warning: increasing threshold to {thresh:.2f}")
            w = np.ones(w.shape)
        else:
            iter = iter - 1

    # restrict covariance estimate to non-artifactual part
    X = demean(X, w)
    c0, _ = tscov(X, None, w)

    # Phase 2
    # -------------------------------------------------------------------------
    # We now know which part contains channel-specific artifacts (w==0 for
    # artifact part), and we have an estimate of the covariance matrix of the
    # artifact-free part.

    # Second eccentricity measure
    d = _eccentricity(X, None, w, n_smooth)

    rank = np.argsort(d, axis=1)[:, ::-1].astype(float)
    rank[np.where(w)[0], :] = np.nan  # exclude parts that are not eccentric

    depth = np.min((depth, n_chans - 1))
    ww = np.ones(X.shape)
    y = X.copy()
    for i_depth in np.arange(depth):
        # Fix each channel by projecting on other channels
        i_fixed = n_chans
        n_fixed = 0
        for ch in np.arange(n_chans):
            neighbours = _closest_neighbours(closest, ch, n_chans)

            # find samples where channel `ch` is the most eccentric
            bad_samples = np.where(ch == rank[:, i_depth])[0]
            if i_depth != 0:  # exclude if not very bad
                bad_samples = np.delete(
                    bad_samples, np.where(d[bad_samples, ch] < thresh)[0])

            n_fixed = n_fixed + np.size(bad_samples)
            if len(bad_samples) == 0:
                i_fixed = i_fixed - 1
                continue

            ww[bad_samples, ch] = 0

            # project this channel on other channels
            z = _project_channel(y[bad_samples, :][:, neighbours], c0, ch,
                                 neighbours)
            y[bad_samples, ch] = z.squeeze()  # fix

        if verbose:
            print(f"depth: {i_depth + 1}")
            print(f"fixed channels: {i_fixed}")
            print(f"fixed samples: {n_fixed}")
            print(f"ratio: {wpwr(X)[0] / p00:.2f}")

    y = demean(y)
    y *= norm
    y += intercept

    # Reset nan and flat channels to zero
    if idx_nan.any():
        y[:, idx_nan] = np.nan
    if idx_zero.any():
        y[:, idx_zero] = 0

    if verbose:
        print(f"power ratio: {wpwr(y)[0] / p0:.2f}")

    if verbose == "debug":
        _diagnostics(X * norm + intercept, y, d, thresh)

    if ndims == 3:  # fold back into trials
        y = fold(y, n_samples)
        w = fold(w, n_samples)
        ww = fold(ww, n_samples)

    return y, w, ww


def _closest_neighbours(closest, ch, n_chans):
    """Find indices of closes neighbours to a given channel."""
    if len(closest) > 0:
        neighbours = closest[ch, :]
    else:
        neighbours = np.delete(np.arange(n_chans), ch)

    # in case closest includes channels not in data
    neighbours = np.delete(neighbours, np.where(neighbours > n_chans)[0])

    return neighbours


def _eccentricity(X, y, weights, n_smooth=0):
    """Compute (smoothed) eccentricity of each time point.

    If y is not defined, the eccentricity measure is based on the absolute
    value of the signal.

    If both X and y are defined, the eccentricity is based on distance between
    X (a given channel's data), and y (the projection from all other channels).

    """
    if weights is None:
        weights = np.ones((X.shape[0], 1))

    if y is None:
        e = np.abs(X) / np.nanstd(X[np.where(weights)[0], :], axis=0)
    else:
        eps = np.finfo(float).eps
        # difference from projection, add eps to avoid error on simulated data
        dx = np.abs(y - X) + eps
        e = dx / np.mean(dx[weights > 0], axis=0)  # eccentricity measure

    # Smooth eccentricity slightly over time
    if n_smooth > 0:
        e = filtfilt(np.ones((n_smooth,)) / n_smooth, 1, e, axis=0)

    return e


def _project_channel(X, c0, ch, neighbours, pca_threshold=1e-15):
    """Compute projection of a channel on other channels."""
    # PCA other channels to remove weak dimensions
    c01 = c0[neighbours, :][:, neighbours]
    topcs, eigenvalues = pca(c01)
    idx = np.where(eigenvalues / np.max(eigenvalues) > pca_threshold)[0]
    topcs = topcs[:, idx]

    # X = c0[ch, neighbours].dot(topcs) / (topcs.T.dot(c01).dot(topcs))
    # In Matlab X = A / B performs mrdivide(), and solves for XB=A
    # In python, the best way to approximate this is with linalg.lstsq
    A = c0[ch, neighbours].dot(topcs)[None, :]
    B = topcs.T.dot(c01).dot(topcs)
    proj = mrdivide(A, B)
    y = X.dot(topcs.dot(proj.T))  # projection

    return y


def _diagnostics(X, y, d, thresh):
    """Draw some diagnostic plots."""
    import matplotlib.pyplot as plt

    # ax1.imshow(w, aspect='auto')

    f, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1)
    ax1.plot(X, lw=.5)
    ax1.set_title("Signal + Artifacts")
    ax1.set_xticklabels([])

    ax2.plot(demean(y), lw=.5)
    ax2.set_title("Denoised")
    ax2.set_xticklabels([])

    ax3.plot(X - demean(y), lw=.5)
    ax3.set_title("Residual")
    ax3.set_xticklabels([])

    ax4.plot(d, lw=.5, alpha=.3)
    d[d < thresh] = None
    ax4.plot(d, lw=1)
    ax4.axhline(thresh, lw=2, color="k", ls=":")
    ax4.set_ylim([0, thresh + 1])
    ax4.set_title("Eccentricity")
    ax4.set_xlabel("samples")

    f.set_layout_engine("tight")
    plt.show()



"""
Rhythmic Entrainment Source Separation (RESS) example
=====================================================

Find the linear combinations of multichannel data that maximize the
signal-to-noise ratio of the narrow-band steady-state response in the frequency
domain.

Uses `meegkit.RESS()`.
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as ss

from meegkit import ress
from meegkit.utils import fold, matmul3d, snr_spectrum, unfold

# import config

rng = np.random.default_rng(9)

###############################################################################
# Create synthetic data
# -----------------------------------------------------------------------------
# Create synthetic data containing a single oscillatory component at 12 hz.
n_times = 1000
n_chans = 10
n_trials = 30
target = 12
sfreq = 250
noise_dim = 8
SNR = .2
t0 = 100

# source
source = np.sin(2 * np.pi * target * np.arange(n_times - t0) / sfreq)[None].T
s = source * rng.standard_normal((1, n_chans))
s = s[:, :, np.newaxis]
s = np.tile(s, (1, 1, n_trials))
signal = np.zeros((n_times, n_chans, n_trials))
signal[t0:, :, :] = s

# noise
noise = np.dot(
    unfold(rng.standard_normal((n_times, noise_dim, n_trials))),
    rng.standard_normal((noise_dim, n_chans)))
noise = fold(noise, n_times)

# mix signal and noise
signal = SNR * signal /  np.sqrt(np.mean(signal ** 2))
noise = noise / np.sqrt(np.mean(noise ** 2))
data = signal + noise

# Plot
f, ax = plt.subplots(3)
ax[0].plot(signal[:, 0, 0], c="C0", label="source")
ax[1].plot(noise[:, 1, 0], c="C1", label="noise")
ax[2].plot(data[:, 1, 0], c="C2", label="mixture")
ax[0].legend()
ax[1].legend()
ax[2].legend()

###############################################################################
# Enhance oscillatory activity using RESS
# -----------------------------------------------------------------------------
# Apply RESS
r = ress.RESS(sfreq=sfreq, peak_freq=target, compute_unmixing=True)
out = r.fit_transform(data)

# Compute PSD
nfft = 250
df = sfreq / nfft  # frequency resolution
bins, psd = ss.welch(np.squeeze(out), sfreq, window="hamming", nperseg=nfft,
                     noverlap=125, axis=0)
psd = psd.mean(axis=1, keepdims=True)  # average over trials
snr = snr_spectrum(psd, bins, skipbins=2, n_avg=2)

f, ax = plt.subplots(1)
ax.plot(bins, snr, "o", label="SNR")
ax.plot(bins[bins == target], snr[bins == target], "ro", label="Target SNR")
ax.axhline(1, ls=":", c="grey", zorder=0)
ax.axvline(target, ls=":", c="grey", zorder=0)
ax.set_ylabel("SNR (a.u.)")
ax.set_xlabel("Frequency (Hz)")
ax.set_xlim([0, 40])

###############################################################################
# Project components back into sensor space to see the effects of RESS on the
# average SSVEP.
fromress = r.from_ress
proj = matmul3d(out, fromress)
f, ax = plt.subplots(n_chans, 2, sharey="col")
for c in range(n_chans):
    ax[c, 0].plot(data[:, c].mean(-1), lw=.5, label="data")
    ax[c, 1].plot(proj[:, c].mean(-1), lw=.5, label="projection")
    ax[c, 0].set_ylabel(f"ch{c}")
    if c < n_chans:
        ax[c, 0].set_xticks([])
        ax[c, 1].set_xticks([])

ax[0, 0].set_title("Trial average (before)")
ax[0, 1].set_title("Trial average (after)")
plt.show()




"""
Sparse time artifact removal on simulated data
==============================================

This example is similar to test_nt_star.m in Noisetools. Results are equivalent
(within numerical precision) to the Matlab code.

"""
import matplotlib.pyplot as plt
import numpy as np

from meegkit import star
from meegkit.utils import demean, normcol

rng = np.random.default_rng(9)

###############################################################################
# Create simulated data
# -----------------------------------------------------------------------------
# Simulated data consist of N channels, 1 sinusoidal target, N-3 noise sources,
# with temporally local artifacts on each channel.

# Create simulated data
nchans = 10
n_samples = 1000
f = 2
target = np.sin(np.arange(n_samples) / n_samples * 2 * np.pi * f)
target = target[:, np.newaxis]
noise = rng.standard_normal((n_samples, nchans - 3))

# Create artifact signal
SNR = np.sqrt(1)
x0 = normcol(np.dot(noise, rng.standard_normal((noise.shape[1], nchans)))) + \
    SNR * target * rng.standard_normal((1, nchans))
x0 = demean(x0)
artifact = np.zeros(x0.shape)
for k in np.arange(nchans):
    artifact[k * 100 + np.arange(20), k] = 1
x = x0 + 10 * artifact

# This is to compare with matlab numerically
# from scipy.io import loadmat
# mat = loadmat('/Users/nicolas/Toolboxes/NoiseTools/TEST/X.mat')
# x = mat['x']
# x0 = mat['x0']

###############################################################################
# Apply STAR
# -----------------------------------------------------------------------------
y, w, _ = star.star(x, 2)

###############################################################################
# Plot results
# -----------------------------------------------------------------------------
f, (ax1, ax2, ax3) = plt.subplots(3, 1)
ax1.plot(x, lw=.5)
ax1.set_title(f"Signal + Artifacts (SNR = {SNR})")
ax2.plot(y, lw=.5)
ax2.set_title("Denoised")
ax3.plot(demean(y) - x0, lw=.5)
ax3.set_title("Residual")
f.set_tight_layout(True)
plt.show()



"""
Ringing artifact reduction example
==================================

This example shows how to subtract the impulse response from a filter to
reduce ringing artifacts.

"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import butter, lfilter

from meegkit.detrend import reduce_ringing

rng = np.random.default_rng(9)

###############################################################################
# Detrending
# =============================================================================

###############################################################################
# Basic example with a linear trend
# -----------------------------------------------------------------------------
# Simulate the effect of filtering a signal containing a discontinuity, and try
# to remove the resulting ringing artifact by subtracing the opposite of the
# impulse response.

x = np.arange(1000) < 1
[b, a] = butter(6, 0.2)     # Butterworth filter design
x = lfilter(b, a, x) * 50    # Filter data using above filter
x = np.roll(x, 500)
x = x[:, None] + rng.standard_normal((1000, 2))
y = reduce_ringing(x, samples=np.array([500]))

plt.figure()
plt.plot(x + np.array([-10, 10]), "C0", label="before")
plt.plot(y + np.array([-10, 10]), "C1:", label="after")
plt.legend()
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import matplotlib as mpl

mpl.use("agg")

curdir = os.path.dirname(__file__)
sys.path.append(os.path.abspath(os.path.join(curdir, "..")))
sys.path.append(os.path.abspath(os.path.join(curdir, "..", "meegkit")))

import meegkit # noqa

# -- Project information -----------------------------------------------------

project = "MEEGkit"
copyright = "2024, Nicolas Barascud"
author = "Nicolas Barascud"
release = meegkit.__version__
version = meegkit.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "numpydoc",
    "jupyter_sphinx",
    "sphinx_gallery.gen_gallery",
    "sphinxemoji.sphinxemoji",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "config.py"]

# generate autosummary even if no references
# autosummary_generate = True
autodoc_default_options = {
    "members": True,
    "special-members": "__init__",
    "undoc-members": True,
    "show-inheritance": True,
    "exclude-members": "__weakref__"
}
numpydoc_show_class_members = False

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_theme_options = {
   "logo": {
      "image_light": "_static/logo.png",
      "image_dark": "_static/logo-dark.png",
      "text": "meegkit",
   },
    "show_toc_level": 1,
    "external_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/nbara/python-meegkit",
        }
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/nbara/python-meegkit",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "Twitter",
            "url": "https://twitter.com/lebababa",
            "icon": "fa-brands fa-twitter",
        },
    ],
    "use_edit_page_button": True,
}

html_context = {
    "github_user": "nbara",
    "github_repo": "python-meegkit",
    "github_version": "master",
    "doc_path": "doc",
}

# -- Options for Sphinx-gallery HTML ------------------------------------------

sphinx_gallery_conf = {
    "doc_module": ("meegkit",),
    "examples_dirs": "../examples",   # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
    "filename_pattern": "/example_",
    "ignore_pattern": "config.py",
    "run_stale_examples": False,
}

suppress_warnings = ["config.cache"]


"""M/EEG denoising utilities in python."""
__version__ = "0.1.9"

from . import asr, cca, detrend, dss, lof, phase, ress, sns, star, trca, tspca, utils

__all__ = ["asr", "cca", "detrend", "dss", "lof", "phase", "ress", "sns", "star",
           "trca", "tspca", "utils"]



import numpy as np
import pytest

from meegkit.utils.buffer import Buffer


def test_push():
    buf = Buffer(32, 1)
    data = np.arange(1000)
    buf.push(data[:10])
    assert buf.counter == 10
    assert buf.head == 10
    assert buf.tail == 0

def test_get_new_samples():
    buf = Buffer(32, 1)
    data = np.arange(1000)
    buf.push(data[:10])
    samples = buf.get_new_samples(5)
    np.testing.assert_array_equal(samples.flatten(), data[:5])
    assert buf.tail == 5
    samples = buf.get_new_samples(5)
    np.testing.assert_array_equal(samples.flatten(), data[5:10])

def test_reset():
    buf = Buffer(32, 1)
    data = np.arange(1000)
    buf.push(data[:10])
    buf.reset()
    assert buf.counter == 0
    assert buf.head == 0
    assert buf.tail == 0

def test_repr():
    buf = Buffer(32, 1)
    repr = buf.__repr__()
    assert repr == "Buffer(32, 1)\n> counter: 0\n> head: 0\n> tail: 0\n"

def test_overflow_warning():
    buf = Buffer(32, 1)
    data = np.arange(1000)
    with pytest.warns(UserWarning, match="Buffer overflow: some old data has been lost"):
        buf.push(data)

def test_repeated_push():
    buf = Buffer(32, 1)
    data = np.arange(1000)

    print(buf)
    buf.push(data[:10])
    print(buf)
    buf.push(data[10:20])
    print(buf)

    with pytest.warns(UserWarning, match="Buffer overflow: some old data has been discarded"):
        buf.push(data[20:40])
        print(buf)
        assert buf.counter == 40
        assert buf.head == 8
        assert buf._head == 40
        assert buf.tail == 8
        assert buf._tail == 8

    out = buf.get_new_samples()
    print(out.flatten())
    np.testing.assert_array_equal(out.flatten(), data[8:40]) # first 8 samples are lost


if __name__ == "__main__":
    # test_get_new_samples()
    # test_repeated_push()
    pytest.main([__file__])


"""
Task-related component analysis for SSVEP detection
===================================================

Sample code for the task-related component analysis (TRCA)-based steady
-state visual evoked potential (SSVEP) detection method [1]_. The filter
bank analysis can also be combined to the TRCA-based algorithm [2]_ [3]_.

This code is based on the Matlab implementation from:
https://github.com/mnakanishi/TRCA-SSVEP

Uses `meegkit.trca.TRCA()`.

References
----------
.. [1] M. Nakanishi, Y. Wang, X. Chen, Y.-T. Wang, X. Gao, and T.-P. Jung,
   "Enhancing detection of SSVEPs for a high-speed brain speller using
   task-related component analysis", IEEE Trans. Biomed. Eng, 65(1): 104-112,
   2018.
.. [2] X. Chen, Y. Wang, S. Gao, T. -P. Jung and X. Gao, "Filter bank
   canonical correlation analysis for implementing a high-speed SSVEP-based
   brain-computer interface", J. Neural Eng., 12: 046008, 2015.
.. [3] X. Chen, Y. Wang, M. Nakanishi, X. Gao, T. -P. Jung, S. Gao,
   "High-speed spelling with a noninvasive brain-computer interface",
   Proc. Int. Natl. Acad. Sci. U. S. A, 112(44): E6058-6067, 2015.

"""
# Authors: Giuseppe Ferraro <giuseppe.ferraro@isae-supaero.fr>
#          Nicolas Barascud <nicolas.barascud@gmail.com>
import os
import time

import matplotlib.pyplot as plt
import numpy as np
import scipy.io

from meegkit.trca import TRCA
from meegkit.utils.trca import itr, normfit, round_half_up

t = time.time()

###############################################################################
# Parameters
# -----------------------------------------------------------------------------
dur_gaze = 0.5  # data length for target identification [s]
delay = 0.13  # visual latency being considered in the analysis [s]
n_bands = 5  # number of sub-bands in filter bank analysis
is_ensemble = True  # True = ensemble TRCA method; False = TRCA method
alpha_ci = 0.05   # 100*(1-alpha_ci): confidence interval for accuracy
sfreq = 250  # sampling rate [Hz]
dur_shift = 0.5  # duration for gaze shifting [s]
list_freqs = np.array(
    [[x + 8.0 for x in range(8)],
     [x + 8.2 for x in range(8)],
     [x + 8.4 for x in range(8)],
     [x + 8.6 for x in range(8)],
     [x + 8.8 for x in range(8)]]).T  # list of stimulus frequencies
n_targets = list_freqs.size  # The number of stimuli

# Useful variables (no need to modify)
dur_gaze_s = round_half_up(dur_gaze * sfreq)  # data length [samples]
delay_s = round_half_up(delay * sfreq)  # visual latency [samples]
dur_sel_s = dur_gaze + dur_shift  # selection time [s]
ci = 100 * (1 - alpha_ci)  # confidence interval

###############################################################################
# Load data
# -----------------------------------------------------------------------------
path = os.path.join("..", "tests", "data", "trcadata.mat")
eeg = scipy.io.loadmat(path)["eeg"]

n_trials, n_chans, n_samples, n_blocks = eeg.shape

# Convert dummy Matlab format to (sample, channels, trials) and construct
# vector of labels
eeg = np.reshape(eeg.transpose([2, 1, 3, 0]),
                 (n_samples, n_chans, n_trials * n_blocks))
labels = np.array([x for x in range(n_targets)] * n_blocks)
crop_data = np.arange(delay_s, delay_s + dur_gaze_s)
eeg = eeg[crop_data]

###############################################################################
# TRCA classification
# -----------------------------------------------------------------------------
# Estimate classification performance with a Leave-One-Block-Out
# cross-validation approach.
#
# To get a sense of the filterbank specification in relation to the stimuli
# we can plot the individual filterbank sub-bands as well as the target
# frequencies (with their expected harmonics in the EEG spectrum). We use the
# filterbank specification described in [2]_.

filterbank = [[(6, 90), (4, 100)],  # passband, stopband freqs [(Wp), (Ws)]
              [(14, 90), (10, 100)],
              [(22, 90), (16, 100)],
              [(30, 90), (24, 100)],
              [(38, 90), (32, 100)],
              [(46, 90), (40, 100)],
              [(54, 90), (48, 100)]]

f, ax = plt.subplots(1, figsize=(7, 4))
for i, _band in enumerate(filterbank):
    ax.axvspan(ymin=i / len(filterbank) + .02,
               ymax=(i + 1) / len(filterbank) - .02,
               xmin=filterbank[i][1][0], xmax=filterbank[i][1][1],
               alpha=0.2, facecolor=f"C{i}")
    ax.axvspan(ymin=i / len(filterbank) + .02,
               ymax=(i + 1) / len(filterbank) - .02,
               xmin=filterbank[i][0][0], xmax=filterbank[i][0][1],
               alpha=0.5, label=f"sub-band{i}", facecolor=f"C{i}")

for f in list_freqs.flat:
    colors = np.ones((9, 4))
    colors[:, :3] = np.linspace(0, .5, 9)[:, None]
    ax.scatter(f * np.arange(1, 10), [f] * 9, c=colors, s=8, zorder=100)

ax.set_ylabel("Stimulus frequency (Hz)")
ax.set_xlabel("EEG response frequency (Hz)")
ax.set_xlim([0, 102])
ax.set_xticks(np.arange(0, 100, 10))
ax.grid(True, ls=":", axis="x")
ax.legend(bbox_to_anchor=(1.05, .5), fontsize="small")
plt.tight_layout()
plt.show()

###############################################################################
# Now perform the TRCA-based SSVEP detection algorithm
trca = TRCA(sfreq, filterbank, is_ensemble)

print("Results of the ensemble TRCA-based method:\n")
accs = np.zeros(n_blocks)
itrs = np.zeros(n_blocks)
for i in range(n_blocks):

    # Select all folds except one for training
    traindata = np.concatenate(
        (eeg[..., :i * n_trials],
         eeg[..., (i + 1) * n_trials:]), 2)
    y_train = np.concatenate(
        (labels[:i * n_trials], labels[(i + 1) * n_trials:]), 0)

    # Construction of the spatial filter and the reference signals
    trca.fit(traindata, y_train)

    # Test stage
    testdata = eeg[..., i * n_trials:(i + 1) * n_trials]
    y_test = labels[i * n_trials:(i + 1) * n_trials]
    estimated = trca.predict(testdata)

    # Evaluation of the performance for this fold (accuracy and ITR)
    is_correct = estimated == y_test
    accs[i] = np.mean(is_correct) * 100
    itrs[i] = itr(n_targets, np.mean(is_correct), dur_sel_s)
    print(f"Block {i}: accuracy = {accs[i]:.1f}, \tITR = {itrs[i]:.1f}")

# Mean accuracy and ITR computation
mu, _, muci, _ = normfit(accs, alpha_ci)
print(f"\nMean accuracy = {mu:.1f}%\t({ci:.0f}% CI: {muci[0]:.1f}-{muci[1]:.1f}%)")  # noqa

mu, _, muci, _ = normfit(itrs, alpha_ci)
print(f"Mean ITR = {mu:.1f}\t({ci:.0f}% CI: {muci[0]:.1f}-{muci[1]:.1f})")
if is_ensemble:
    ensemble = "ensemble TRCA-based method"
else:
    ensemble = "TRCA-based method"

print(f"\nElapsed time: {time.time()-t:.1f} seconds")



"""Utility functions."""
from .auditory import AuditoryFilterbank, GammatoneFilterbank, erb2hz, erbspace, hz2erb
from .base import mldivide, mrdivide
from .coherence import (
    cross_coherence,
    plot_polycoherence,
    plot_polycoherence_1d,
    plot_signal,
    polycoherence_0d,
    polycoherence_1d,
    polycoherence_1d_sum,
    polycoherence_2d,
)
from .covariances import (
    block_covariance,
    convmtx,
    cov_lags,
    nonlinear_eigenspace,
    pca,
    regcov,
    tscov,
    tsxcov,
)
from .denoise import (
    demean,
    find_outlier_samples,
    find_outlier_trials,
    mean_over_trials,
    wpwr,
)
from .matrix import (
    fold,
    matmul3d,
    multishift,
    multismooth,
    normcol,
    relshift,
    shift,
    shiftnd,
    sliding_window,
    theshapeof,
    unfold,
    unsqueeze,
    widen_mask,
)
from .sig import (
    gaussfilt,
    hilbert_envelope,
    slope_sum,
    smooth,
    spectral_envelope,
    teager_kaiser,
)
from .stats import (
    bootstrap_ci,
    bootstrap_snr,
    cronbach,
    rms,
    robust_mean,
    rolling_corr,
    snr_spectrum,
)
from .testing import create_line_data



"""TRCA tests."""
import os

import numpy as np
import pytest
import scipy.io

from meegkit.trca import TRCA
from meegkit.utils.trca import itr, normfit, round_half_up

##########################################################################
# Load data
# -----------------------------------------------------------------------------
path = os.path.join(".", "tests", "data", "trcadata.mat")
mat = scipy.io.loadmat(path)
eeg = mat["eeg"]

n_trials = eeg.shape[0]
n_chans = eeg.shape[1]
n_samples = eeg.shape[2]
n_blocks = eeg.shape[3]

list_freqs = np.concatenate(
    [[x + 8 for x in range(8)],
        [x + 8.2 for x in range(8)],
        [x + 8.4 for x in range(8)],
        [x + 8.6 for x in range(8)],
        [x + 8.8 for x in range(8)]])  # list of stimulus frequencies
n_targets = len(list_freqs)  # The number of stimuli

# Convert dummy Matlab format to (sample, channels, trials) and construct
# vector of labels
eeg = np.reshape(eeg.transpose([2, 1, 3, 0]),
                 (n_samples, n_chans, n_trials * n_blocks))
labels = np.array([x for x in range(n_targets)] * n_blocks)

# We use the filterbank specification described in [2]_.
filterbank = [[[6, 90], [4, 100]],  # passband freqs, stopband freqs (Wp, Ws)
              [[14, 90], [10, 100]],
              [[22, 90], [16, 100]],
              [[30, 90], [24, 100]],
              [[38, 90], [32, 100]]]


@pytest.mark.parametrize("ensemble", [True, False])
@pytest.mark.parametrize("method", ["original", "riemann"])
@pytest.mark.parametrize("regularization", ["schaefer", "scm"])
def test_trca(ensemble, method, regularization):
    """Test TRCA."""
    if method == "original" and regularization == "schaefer":
        pytest.skip("regularization only used for riemann version")

    len_gaze_s = 0.5  # data length for target identification [s]
    len_delay_s = 0.13  # visual latency being considered in the analysis [s]
    alpha_ci = 0.05   # 100*(1-alpha_ci): confidence interval for accuracy
    sfreq = 250  # sampling rate [Hz]
    len_shift_s = 0.5  # duration for gaze shifting [s]

    # useful variables
    len_gaze_smpl = round_half_up(len_gaze_s * sfreq)  # data length [samples]
    len_delay_smpl = round_half_up(len_delay_s * sfreq)  # visual latency [samples]
    len_sel_s = len_gaze_s + len_shift_s  # selection time [s]
    ci = 100 * (1 - alpha_ci)  # confidence interval

    crop_data = np.arange(len_delay_smpl, len_delay_smpl + len_gaze_smpl)

    ##########################################################################
    # TRCA classification
    # -----------------------------------------------------------------------------
    # Estimate classification performance with a Leave-One-Block-Out
    # cross-validation approach
    trca = TRCA(sfreq, filterbank, ensemble=ensemble, method=method,
                estimator=regularization)
    accs = np.zeros(2)
    itrs = np.zeros(2)
    for i in range(2):

        # Training stage
        traindata = eeg.copy()[crop_data]

        # Select all folds except one for training
        traindata = np.concatenate(
            (traindata[..., :i * n_trials],
             traindata[..., (i + 1) * n_trials:]), 2)
        y_train = np.concatenate(
            (labels[:i * n_trials], labels[(i + 1) * n_trials:]), 0)

        # Construction of the spatial filter and the reference signals
        trca.fit(traindata, y_train)

        # Test stage
        testdata = eeg[crop_data, :, i * n_trials:(i + 1) * n_trials]
        y_test = labels[i * n_trials:(i + 1) * n_trials]
        estimated = trca.predict(testdata)

        # Evaluation of the performance for this fold (accuracy and ITR)
        is_correct = estimated == y_test
        accs[i] = np.mean(is_correct) * 100
        itrs[i] = itr(n_targets, np.mean(is_correct), len_sel_s)
        print(f"Block {i}: accuracy = {accs[i]:.1f}, \tITR = {itrs[i]:.1f}")

    # Mean accuracy and ITR computation
    mu, _, muci, _ = normfit(accs, alpha_ci)
    print(f"Mean accuracy = {mu:.1f}%\t({ci:.0f}% CI: {muci[0]:.1f}-{muci[1]:.1f}%)")  # noqa
    if method != "riemann" or (regularization == "scm" and ensemble):
        assert mu > 95
    mu, _, muci, _ = normfit(itrs, alpha_ci)
    print(f"Mean ITR = {mu:.1f}\t({ci:.0f}% CI: {muci[0]:.1f}-{muci[1]:.1f}%)")
    if method != "riemann" or (regularization == "scm" and ensemble):
        assert mu > 300


if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_trcacode()



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
from scipy.io import loadmat
from sklearn.cross_decomposition import CCA

from meegkit.cca import cca_crossvalidate, mcca, nt_cca
from meegkit.utils import multishift, tscov


def test_cca():
    """Test CCA."""
    # Compare results with Matlab
    # x = rng.randn(1000, 11)
    # y = rng.randn(1000, 9)
    # x = demean(x).squeeze()
    # y = demean(y).squeeze()
    mat = loadmat("./tests/data/ccadata.mat")
    x = mat["x"]
    y = mat["y"]
    A2 = mat["A2"]
    B2 = mat["B2"]

    A1, B1, R = nt_cca(x, y)  # if mean(A1(:).*A2(:))<0; A2=-A2; end
    X1 = np.dot(x, A1)
    Y1 = np.dot(y, B1)
    C1 = tscov(np.hstack((X1, Y1)))[0]

    # Sklearn CCA
    cca = CCA(n_components=9, scale=False, max_iter=int(1e6))
    X2, Y2 = cca.fit_transform(x, y)
    # C2 = tscov(np.hstack((X2, Y2)).T)[0]
    # import matplotlib.pyplot as plt
    # f, (ax1, ax2) = plt.subplots(2, 1)
    # ax1.imshow(C1)
    # ax2.imshow(C2)
    # plt.show()
    # assert_almost_equal(C1, C2, decimal=4)

    # Compare with matlab
    X2 = np.dot(x, A2)
    Y2 = np.dot(y, B2)
    C2 = tscov(np.hstack((X2, Y2)))[0]

    assert_almost_equal(C1, C2)


def test_cca2():
    """Simulate correlations."""
    # import matplotlib.pyplot as plt
    rng = np.random.RandomState(2022)
    x = rng.randn(10000, 20)
    y = rng.randn(10000, 8)
    y[:, :2] = x[:, :2]
    # perfectly correlated
    y[:, 2:4] = x[:, 2:4] + rng.randn(10000, 2)
    #  1/2 correlated
    y[:, 4:6] = x[:, 4:6] + rng.randn(10000, 2) * 3
    #  1/4 correlated
    y[:, 6:8] = rng.randn(10000, 2)
    # uncorrelated
    [A, B, R] = nt_cca(x, y)

    assert_almost_equal(np.dot(np.dot(x, A).T, np.dot(x, A)), np.eye(8))
    assert_almost_equal(np.dot(np.dot(y, B).T, np.dot(y, B)), np.eye(8))

    # f, ax = plt.subplots(3, 2)
    # ax[0, 0].imshow(A, aspect='auto')
    # ax[0, 0].set_title('A')
    # ax[0, 1].imshow(B, aspect='auto')
    # ax[0, 1].set_title('B')
    # ax[1, 0].plot(R, '.-')
    # ax[1, 0].set_title('R')
    # ax[1, 1].imshow(np.dot(np.dot(x, A).T, np.dot(x, A)), aspect='auto')
    # ax[1, 1].set_title('covariance of x * A')
    # ax[2, 0].imshow(np.dot(np.dot(y, B).T, np.dot(y, B)), aspect='auto')
    # ax[2, 0].set_title('covariance of y * B')
    # f.set_tight_layout(True)
    # plt.show()


def test_cca_scaling():
    """Test CCA with MEG data."""
    data = np.load("./tests/data/ccadata_meg_2trials.npz")
    raw = data["arr_0"]
    env = data["arr_1"]

    # Test with scaling (unit: fT)
    A0, B0, R0 = nt_cca(raw * 1e15, env)

    # Test without scaling (unit: T)
    A1, B1, R1 = nt_cca(raw, env)

    np.testing.assert_almost_equal(R0, R1)


def test_canoncorr():
    """Compare with Matlab's canoncorr."""
    x = np.array([[16, 2, 3, 13],
                  [5, 11, 10, 8],
                  [9, 7, 6, 12],
                  [4, 14, 15, 1]])
    y = np.array([[1, 5, 9],
                  [2, 6, 10],
                  [3, 7, 11],
                  [4, 8, 12]])
    [A, B, R] = nt_cca(x, y)
    canoncorr = np.array([[-0.5000, -0.6708],
                          [-0.5000, 0.2236],
                          [-0.5000, -0.2236],
                          [-0.5000, 0.6708]])

    # close enough
    assert_almost_equal(np.cov(np.dot(x, A).T), np.cov(canoncorr.T), decimal=4)


def test_correlated():
    """Test x & y perfectly correlated."""
    rng = np.random.RandomState(2022)
    x = rng.randn(1000, 10)
    y = rng.randn(1000, 10)

    y = x[:, rng.permutation(10)]  # +0.000001*y;

    [A1, B1, R1] = nt_cca(x, y)

    C = tscov(np.hstack((np.dot(x, A1), np.dot(y, B1))))[0]

    # import matplotlib.pyplot as plt
    # f, ax1 = plt.subplots(1, 1)
    # im = ax1.imshow(C)
    # plt.colorbar(im)
    # plt.show()

    for i in np.arange(C.shape[0]):
        assert_almost_equal(C[i, i], C[i, (i + 10) % 10], decimal=4)
        assert_almost_equal(C[i, i], C[i, (i - 10) % 10], decimal=4)


def test_cca_lags():
    """Test multiple lags."""
    mat = loadmat("./tests/data/ccadata.mat")
    x = mat["x"]
    y = mat["y"]
    y[:, :3] = x[:, :3]
    lags = np.arange(-10, 11, 1)
    A1, B1, R1 = nt_cca(x, y, lags)

    assert A1.ndim == B1.ndim == 3
    assert A1.shape[-1] == B1.shape[-1] == lags.size

    # import matplotlib.pyplot as plt
    # f, ax1 = plt.subplots(1, 1)
    # ax1.plot(lags, R1.T)
    # plt.show()

    for c in np.arange(3):  # test first 3 components have peaks at 0
        assert_equal(lags[np.argmax(R1[c, :])], 0)


def test_cca_crossvalidate():
    """Test CCA with crossvalidation."""
    rng = np.random.RandomState(2023)
    # x = rng.randn(1000, 11)
    # y = rng.randn(1000, 9)
    # xx = [x, x, x]
    # yy = [x[:, :9], y, y]

    mat = loadmat("./tests/data/ccadata2.mat")
    xx = mat["x"]
    yy = mat["y"]
    R1 = mat["R"]  # no shifts

    # Test with no shifts
    A, B, R = cca_crossvalidate(xx, yy)

    assert_almost_equal(R, R1, decimal=2)

    # Create data where 1st comps should be uncorrelated, and 2nd and 3rd comps
    # are very correlated
    x = rng.randn(1000, 10)
    y = rng.randn(1000, 10)
    xx = [x, x, x]
    yy = [y, x, x]
    A, B, R = cca_crossvalidate(xx, yy)
    assert_almost_equal(R[:, :, 0], 0, decimal=1)
    assert_almost_equal(R[:, :, 1:], 1, decimal=1)


def test_cca_crossvalidate_shifts():
    """Test CCA crossvalidation with shifts."""
    rng = np.random.RandomState(2021)
    n_times, n_trials = 10000, 2
    x = rng.randn(n_times, 20, n_trials)
    y = rng.randn(n_times, 8, n_trials)
    # perfectly correlated
    y[:, :2, :] = x[:, :2, :]
    # 1/2 correlated
    y[:, 2:4, :] = x[:, 2:4, :] + rng.randn(n_times, 2, n_trials)
    # 1/4 correlated
    y[:, 4:6, :] = x[:, 4:6, :] + rng.randn(n_times, 2, n_trials) * 3
    # uncorrelated
    y[:, 6:8, :] = rng.randn(n_times, 2, n_trials)

    xx = multishift(x, -np.arange(1, 4), reshape=True, solution="valid")
    yy = multishift(y, -np.arange(1, 4), reshape=True, solution="valid")

    # Test with shifts
    A, B, R = cca_crossvalidate(xx, yy, shifts=[-3, -2, -1, 0, 1, 2, 3])

    # import matplotlib.pyplot as plt
    # f, ax = plt.subplots(n_trials, 1)
    # for i in range(n_trials):
    #     ax[i].plot(R[:, :, i].T)
    # f.set_tight_layout(True)
    # plt.show()


def test_cca_crossvalidate_shifts2():
    """Test CCA crossvalidation with shifts."""
    mat = loadmat("./tests/data/ccacrossdata.mat")
    xx = mat["xx2"]
    yy = mat["yy2"]
    R2 = mat["R"][:, ::-1, :]  # shifts go in reverse direction in noisetools

    # Test with shifts
    A, B, R = cca_crossvalidate(xx, yy, shifts=[-3, -2, -1, 0, 1, 2, 3])

    # correlations are ~ those of Matlab
    assert_almost_equal(R, R2, decimal=3)

    # import matplotlib.pyplot as plt
    # n_trials = xx.shape[-1]
    # f, ax = plt.subplots(n_trials, 2)
    # for i in range(n_trials):
    #     ax[i, 0].plot(R[:, :, i].T)
    #     ax[i, 1].plot(R2[:, :, i].T)
    # f.set_tight_layout(True)
    # plt.show()


def test_mcca(show=False):
    """Test multiway CCA."""
    rng = np.random.RandomState(2021)
    # We create 3 uncorrelated data sets. There should be no common structure
    # between them.

    # Build data
    x1 = rng.randn(10000, 10)
    x2 = rng.randn(10000, 10)
    x3 = rng.randn(10000, 10)
    x = np.hstack((x1, x2, x3))
    C = np.dot(x.T, x)

    ###############################################################################
    # Apply CCA
    [A, score, AA] = mcca(C, 10)
    z = x @ A

    ###############################################################################
    # Plot results
    if show:
        import matplotlib.pyplot as plt
        f, axes = plt.subplots(2, 3, figsize=(10, 6))
        axes[0, 0].imshow(A, aspect="auto")
        axes[0, 0].set_title("mCCA transform matrix")
        axes[0, 1].imshow(A.T @ C @ A, aspect="auto")
        axes[0, 1].set_title("Covariance of\ntransformed data")
        axes[0, 2].imshow(x.T @ x @ A, aspect="auto")
        axes[0, 2].set_title("Cross-correlation between\nraw & transformed data")
        axes[0, 2].set_xlabel("transformed")
        axes[0, 2].set_ylabel("raw")
        ax = plt.subplot2grid((2, 3), (1, 0), colspan=3)
        ax.plot(np.mean(z ** 2, axis=0), ":o")
        ax.set_ylabel("Power")
        ax.set_xlabel("CC")
        plt.tight_layout()
        plt.show()

    # assert np.diag_indices

    # Second example
    # -------------------------------------------------------------------------
    # Now Create 3 data sets with some shared parts.

    # Build data
    x1 = rng.randn(10000, 5)
    x2 = rng.randn(10000, 5)
    x3 = rng.randn(10000, 5)
    x4 = rng.randn(10000, 5)
    x = np.hstack((x2, x1, x3, x1, x4, x1))
    C = np.dot(x.T, x)

    # Apply mCCA
    A, score, AA = mcca(C, 10)
    z = x @ A

    if show:
        f, axes = plt.subplots(2, 3, figsize=(10, 6))
        axes[0, 0].imshow(A, aspect="auto")
        axes[0, 0].set_title("mCCA transform matrix")
        axes[0, 1].imshow(A.T.dot(C.dot(A)), aspect="auto")
        axes[0, 1].set_title("Covariance of\ntransformed data")
        axes[0, 2].imshow(x.T.dot(x.dot(A)), aspect="auto")
        axes[0, 2].set_title("Cross-correlation between\nraw & transformed data")
        axes[0, 2].set_xlabel("transformed")
        axes[0, 2].set_ylabel("raw")
        ax = plt.subplot2grid((2, 3), (1, 0), colspan=3)
        ax.plot(np.mean(z ** 2, axis=0), ":o")
        ax.set_ylabel("Power")
        ax.set_xlabel("CC")
        plt.tight_layout()
        plt.show()

    # Third example
    # -------------------------------------------------------------------------
    # Finally let's create 3 identical 10-channel data sets. Only 10 worthwhile
    # components should be found, and the transformed dataset should perfectly
    # explain all the variance (empty last two block-columns in the
    # cross-correlation plot).

    # Build data
    x1 = rng.randn(10000, 10)
    x = np.hstack((x1, x1, x1))
    C = np.dot(x.T, x)

    # Compute mCCA
    A, score, AA = mcca(C, 10)
    z = x @ A

    # Plot results
    if show:
        f, axes = plt.subplots(2, 3, figsize=(10, 6))
        axes[0, 0].imshow(A, aspect="auto")
        axes[0, 0].set_title("mCCA transform matrix")

        axes[0, 1].imshow(A.T @ C @ A, aspect="auto")
        axes[0, 1].set_title("Covariance of\ntransformed data")

        axes[0, 2].imshow(x.T @ x @ A, aspect="auto")
        axes[0, 2].set_title("Cross-correlation between\nraw & transformed data")
        axes[0, 2].set_xlabel("transformed")
        axes[0, 2].set_ylabel("raw")
        ax = plt.subplot2grid((2, 3), (1, 0), colspan=3)
        ax.plot(np.mean(z ** 2, axis=0), ":o")
        ax.set_ylabel("Power")
        ax.set_xlabel("CC")
        plt.tight_layout()
        plt.show()

    # Only first 10 components should be non-negligible
    diagonal = np.diag(x.T @ x @ A) ** 2
    assert np.all(diagonal[:10] > 1), diagonal[:10]
    assert np.all(diagonal[10:] < .01)

if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_mcca(False)



"""Matrix operation utility functions."""
import warnings

import numpy as np
from numpy.lib.stride_tricks import as_strided


def sliding_window(data, window, step=1, axis=-1, copy=True):
    """Calculate a sliding window over a signal.

    Parameters
    ----------
    data : array
        The array to be slided over.
    window : int
        The sliding window size.
    step : int
        The sliding window stepsize (default=1).
    axis : int
        The axis to slide over (defaults=-1).
    copy : bool
        Return strided array as copy to avoid sideffects when manipulating the
        output array.

    Returns
    -------
    data :  array, shape=(..., n_windows, window_size)
        A matrix whose last dimension corresponds to the window size, and the
        second-to-last dimension corresponds to the number of slices.

    Notes
    -----
    Be wary of setting `copy` to `False` as undesired side effects with the
    output values may occur.

    Examples
    --------
    >>> a = numpy.array([1, 2, 3, 4, 5])
    >>> sliding_window(a, size=3)
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])

    >>> sliding_window(a, size=3, stepsize=2)
    array([[1, 2, 3],
           [3, 4, 5]])

    """
    if axis >= data.ndim:
        raise ValueError("Axis value out of range")
    if step < 1:
        raise ValueError("Stepsize may not be zero or negative")
    if window > data.shape[axis]:
        print("Sliding window size exceeds size of selected axis")
        return data[..., None]

    shape = list(data.shape)
    shape[axis] = np.floor(
        data.shape[axis] / step - window / step + 1).astype(int)
    shape.append(window)

    strides = list(data.strides)
    strides[axis] *= step
    strides.append(data.strides[axis])
    strided = as_strided(data, shape=shape, strides=strides)

    if copy:
        return strided.copy()
    else:
        return strided


def widen_mask(mask, widen=4, axis=0):
    """Widen each 'bad' section of a binary mask by `n` cells.

    Parameters
    ----------
    mask : array
        Masking array. If an element of the mask is True (or 1), the
        corresponding element of the associated array is masked (marked as
        invalid).
    widen : int
        Number of cells to widen mask by.
    axis : int
        Axis to operate on.

    Returns
    -------
    out : array
        Widened mask, of same shape as input mask.

    Examples
    --------
    >>> test = widen_mask(np.array([False, False, False, True, False], 1)
    >>> print(test)
    [False False False True True]]

    """
    dims = mask.ndim
    dtype = mask.dtype

    if axis > dims - 1:
        raise AttributeError("Invalid `axis` value.")

    if widen < 0:
        # This places the desired axis at the front of the shape tuple, then
        # reverses that first axis, and then returns it to its original
        # position.
        if dims > 1:
            mask = np.swapaxes(np.swapaxes(mask, 0, axis)[::-1], 0, axis)
        else:
            mask = mask[::-1]

    def repeat_or(a, n=1):
        """Take a mask and applies logical OR to shifted versions of itself."""
        m = a.copy().astype(bool)
        k = m.copy()
        if n == 0:
            return k
        n = n + 1  # trick so that n=1 means "expand mask by one"

        # lenM and lenK : how many subsequent Trues there are at least
        lenM, lenK = 1, 1

        # run until a combination of both masks has n or more subsequent Trues
        while lenM + lenK < n:
            # append what we have in k to the end of what we have in m
            m[lenM:] |= k[:-lenM]
            m, k = k, m  # swap so that m is again the small one

            # update the lengths
            lenM, lenK = lenK, lenM + lenK

        # see how much m has to be shifted in order to append the missing Trues
        k[n - lenM:] |= m[:-n + lenM]

        return k

    if dims > 1:  # apply along given axis
        out = np.apply_along_axis(repeat_or, axis, mask, n=np.abs(widen))
    else:
        out = repeat_or(mask, n=np.abs(widen))

    if widen < 0:  # un-reverse axis
        if dims > 1:
            out = np.swapaxes(np.swapaxes(out, 0, axis)[::-1], 0, axis)
        else:
            out = out[::-1]

    return out.astype(dtype)


def relshift(X, ref, shifts, fill_value=0, axis=0):
    """Create shifted versions of X relative to ref with padding.

    `ref` is replicated to have the same shape as `X` and padded accordingly.

    Parameters
    ----------
    X : array, shape=(n_samples[, n_epochs][, n_trials])
        Array to shift.
    ref : array, shape=(n_samples[, n_epochs][, n_trials])
        Reference array against which `X` is shifted.
    shifts : array | int
        Array of shifts. Positive shifts mean that X is 'delayed' in time (i.e.
        `y[shift] = X[0]`). Conversely, a negative shift means that X is
        'advanced' (i.e. y[0] =
    fill_value : float
        Value to pad output axis by.
    axis : int
        The axis along which elements are shifted.

    Returns
    -------
    y : array, shape=(n_samples[, n_epochs][, n_trials], n_shifts)
        Shifted array.
    y_ref : array, shape=(n_samples[, n_epochs][, n_trials], n_shifts)
        Reference array, repeated to match `y.shape`. Padding matches that of
        `y`.

    See Also
    --------
    multishift, shift, shiftnd

    """
    shifts, n_shifts = _check_shifts(shifts)
    X = _check_data(X)
    ref = _check_data(ref)

    if X.shape[0] != ref.shape[0]:
        raise AttributeError("X and ref must have same n_times")

    # First we delay X
    y = multishift(X, shifts=shifts, axis=axis, fill_value=fill_value)

    # Then we create as many copies of ref as there are lags
    y_ref = multishift(ref, shifts=np.zeros(n_shifts), axis=axis)

    # We need to find out the indices of the padded values in `y`. For this we
    # use a hack where we feed in an array of ones to multishift(), with a
    # known `fill_value`.
    temp = multishift(np.ones_like(ref), shifts=shifts, axis=axis,
                      fill_value=0)
    mask = temp == 0
    if mask.any():
        y_ref[mask] = fill_value

    return y, y_ref


def multishift(X, shifts, fill_value=0, axis=0, keep_dims=False,
               reshape=False, solution="full"):
    """Apply several shifts along specified axis.

    If `shifts` has multiple values, the output will contain one shift per
    page. Shifted data are padded with `fill_value`.

    Parameters
    ----------
    X : array, shape=(n_samples[, n_chans][, n_trials])
        Array to shift.
    shifts : array, shape=(n_shifts,)
        Array of shifts. Positive shifts mean that X is 'delayed' in time (i.e.
        `y[shift] = X[0]`). Conversely, a negative shift means that X is
        'advanced' (i.e. y[0] =
    fill_value : float | np.nan
        Value to pad output axis by.
    axis : int, optional
        The axis along which elements are shifted.
    keep_dims : bool
        If True, keep singleton dimensions in output.
    reshape : bool
        If True, concatenate channels and lags, yielding an array of shape
        (n_samples, n_chans*n_shifts[, n_trials])
    solution : {'valid', 'full'}
        If `valid`, the output's is cropped along `axis` by `n_shifts` in order
        to remove edge artifacts. If `full`, the output has the same size as X.

    Returns
    -------
    y : array, shape=(n_samples[, n_chans][, n_trials], n_shifts)
        Shifted array.

    See Also
    --------
    relshift, shift, shiftnd

    """
    shifts, n_shifts = _check_shifts(shifts)
    X = _check_data(X)
    n_samples, n_chans, n_trials = theshapeof(X)

    if np.array_equal(shifts, [0]):
        return X.copy()

    # Loop over shifts
    y = np.zeros(X.shape + (n_shifts,))
    for i, s in enumerate(shifts):
        y[..., i] = shift(X, shift=s, fill_value=fill_value, axis=axis)

    if reshape is True:
        if X.ndim == 3:  # n_samples, n_chans, n_trials, n_shifts
            y = np.swapaxes(y, 2, 3)
            y = np.reshape(y, (n_samples, n_chans * n_shifts, n_trials))
        elif X.ndim == 2:  # n_samples, n_chans, n_shifts
            y = np.reshape(y, (n_samples, n_chans * n_shifts))

    if n_shifts == 1 and not keep_dims:
        y = np.squeeze(y, axis=-1)

    if solution == "valid":
        max_neg_shift = np.abs(np.min(np.min(shifts), 0))
        max_pos_shift = np.max((np.max(shifts), 0))
        y = y[max_pos_shift:-max_neg_shift, ...]

    return y


def multismooth(X, smooths, axis=0, keep_dims=False):
    """Apply several shifts along specified axis.

    If `shifts` has multiple values, the output will contain one shift per
    trial. Shifted data are padded with `fill_value`.

    Parameters
    ----------
    X : array, shape=(n_samples[, n_epochs][, n_trials])
        Array to shift.
    smooths : array
        Array of smoothing values (in samples).
    axis : int, optional
        The axis along which elements are shifted (default=0).
    keep_dims : bool
        If True, keep singleton dimensions in output.

    Returns
    -------
    y : array, shape=(n_samples[, n_epochs][, n_trials], n_shifts)
        Shifted array.

    See Also
    --------
    multishift, smooth

    """
    from .sig import smooth

    smooths, n_smooths = _check_shifts(smooths, allow_floats=True)
    X = _check_data(X)

    # Loop over shifts
    y = np.zeros(X.shape + (n_smooths,))
    for i, s in enumerate(smooths):
        y[..., i] = smooth(X, window_len=s, axis=axis)

    if n_smooths == 1 and not keep_dims:
        y = np.squeeze(y, axis=-1)

    return y


def shift(X, shift, fill_value=0, axis=0):
    """Shift array along its first, second or last dimension.

    Output is padded by `fill_value`.

    Parameters
    ----------
    X : array, shape=(n_samples[, n_epochs][, n_trials])
        Multidimensional input array.
    shift : int
        The number of places by which elements are shifted along axis. Positive
        shifts mean that X is 'delayed' in time (i.e. `y[shift] = X[0]`).
        Conversely, a negative shift means that X is 'advanced' (i.e. y[0] =
        X[shift]).
    fill_value : float
        Value to pad output axis by.
    axis : int, optional
        The axis along which elements are shifted.

    Returns
    -------
    y : array
        Output array, with the same shape as `X`.

    See Also
    --------
    relshift, multishift, shiftnd

    """
    if not np.equal(np.mod(shift, 1), 0):
        raise AttributeError("shift must be a single int")

    # reallocate empty array and assign slice.
    y = np.empty_like(X)

    if shift == 0:
        y[:] = X
    else:
        if axis == 0:
            if shift > 0:
                y[:shift, ...] = fill_value
                y[shift:, ...] = X[:-shift]
            elif shift < 0:
                y[shift:,  ...] = fill_value
                y[:shift,  ...] = X[-shift:]

        elif axis == 1:
            if shift > 0:
                y[:, :shift, ...] = fill_value
                y[:, shift:, ...] = X[:, :-shift]
            elif shift < 0:
                y[:, shift:,  ...] = fill_value
                y[:, :shift,  ...] = X[:, -shift:]

        elif axis == -1:
            if shift > 0:
                y[..., :shift] = fill_value
                y[..., shift:] = X[..., :-shift]
            elif shift < 0:
                y[..., shift:] = fill_value
                y[..., :shift] = X[..., -shift:]

        else:
            raise NotImplementedError("Axis must be 0, 1 or -1.")

    return y


def shiftnd(X, shift, fill_value=0, axis=None):
    """Roll array elements along a given axis with padding.

    Elements off the end of the array are treated as zeros. This function is
    slower than function:`shift`, so prefer the latter when possible.

    Parameters
    ----------
    X : array
        Multidimensional input array.
    shift : int
        The number of places by which elements are shifted along axis.
    fill_value : float
        Value to pad output axis by.
    axis : int, optional
        The axis along which elements are shifted. By default, the array is
        flattened before shifting, after which the original shape is restored.

    Returns
    -------
    y : array, (n_samples, [n_epochs, ][n_trials, ])
        Output array, with the same shape as `X`.

    See Also
    --------
    np.roll     : Elements that roll off one end come back on the other.
    np.rollaxis : Roll the specified axis backwards, until it lies in a given
                  position.

    Examples
    --------
    >>> x = np.arange(10)
    >>> shiftnd(x, 2)
    array([0, 0, 0, 1, 2, 3, 4, 5, 6, 7])
    >>> x2 = np.reshape(x, (2,5))
    >>> x2
    array([[0, 1, 2, 3, 4],
           [5, 6, 7, 8, 9]])
    >>> shiftnd(x2, 1)
    array([[0, 0, 1, 2, 3],
           [4, 5, 6, 7, 8]])
    >>> shiftnd(x2, -2)
    array([[2, 3, 4, 5, 6],
           [7, 8, 9, 0, 0]])
    >>> shiftnd(x2, 1, axis=0)
    array([[0, 0, 0, 0, 0],
           [0, 1, 2, 3, 4]])
    >>> shiftnd(x2, -1, axis=0)
    array([[5, 6, 7, 8, 9],
           [0, 0, 0, 0, 0]])
    >>> shiftnd(x2, 1, axis=1)
    array([[0, 0, 1, 2, 3],
           [0, 5, 6, 7, 8]])

    """
    X = np.asanyarray(X)
    if shift == 0:
        return X

    if axis is None:
        n = X.size
        reshape = True
    else:
        n = X.shape[axis]
        reshape = False

    if np.abs(shift) > n:
        y = np.ones_like(X) * fill_value
    elif shift < 0:
        shift += n
        pad = np.ones_like(X.take(np.arange(n - shift), axis)) * fill_value
        y = np.concatenate((X.take(np.arange(n - shift, n), axis), pad), axis)
    else:
        pad = np.ones_like(X.take(np.arange(n - shift, n), axis)) * fill_value
        y = np.concatenate((pad, X.take(np.arange(n - shift), axis)), axis)

    if reshape:
        return y.reshape(X.shape)
    else:
        return y


def theshapeof(X):
    """Return the shape of X."""
    X = _check_data(X)
    # if not isinstance(X, np.ndarray):
    #     raise AttributeError('X must be a numpy array')

    if X.ndim == 3:
        return X.shape[0], X.shape[1], X.shape[2]
    elif X.ndim == 2:
        return X.shape[0], X.shape[1], 1
    elif X.ndim == 1:
        return X.shape[0], 1, 1
    else:
        raise ValueError("Array contains more than 3 dimensions")


def unsqueeze(X):
    """Append singleton dimensions to an array."""
    X = _check_data(X)
    if X.shape != theshapeof(X):
        return X.reshape(theshapeof(X))
    else:
        return X


def fold(X, epoch_size):
    """Fold 2D (n_times, n_channels) X into 3D (n_times, n_chans, n_trials)."""
    if X.ndim == 1:
        X = X[:, np.newaxis]
    if X.ndim > 2:
        raise AttributeError("X must be 2D at most")

    nt = X.shape[0] // epoch_size
    nc = X.shape[1]
    if X.shape[0] / epoch_size >= 1:
        return X.reshape((epoch_size, nt, nc), order="F").transpose([0, 2, 1])
    else:
        return X


def unfold(X):
    """Unfold 3D X into 2D (concatenate trials)."""
    n_samples, n_chans, n_trials = theshapeof(X)
    if X.size == 0:
        return X

    if X.shape == (n_samples,):
        X = X[:, None]

    if n_trials > 1 or X.ndim == 3:
        return np.reshape(
            np.transpose(X, (0, 2, 1)),
            (n_samples * n_trials, n_chans), order="F")
    else:
        return X


def normcol(X, weights=None, return_norm=False):
    """Normalize each column so that its weighted mean square value is 1.

    If X is 3D, pages are concatenated vertically before calculating the
    norm.

    Weight should be either a column vector, or a matrix (2D or 3D) of same
    size as X.

    Parameters
    ----------
    X : array
        X to normalize.
    weights : array
        Weights.
    return_norm : bool
        If True, also return norm vector.

    Returns
    -------
    X_norm : array
        Normalized X.
    norm : array
        Norm.

    """
    if X.ndim == 3:
        n_samples, n_chans, n_trials = theshapeof(X)
        weights = _check_weights(weights, X)
        X = unfold(X)
        if not weights.any():  # no weights
            X_norm, N = normcol(X, return_norm=True)
            N = N ** 2
            X_norm = fold(X_norm, n_samples)

        else:  # weights
            if weights.shape[0] != n_samples:
                raise ValueError("Weight array should have same number of' \
                                 'columns as X")

            if weights.ndim == 2 and weights.shape[1] == 1:
                weights = np.tile(weights, (1, n_samples, n_trials))

            weights = unfold(weights)
            X_norm, N = normcol(X, weights, return_norm=True)
            N = N ** 2
            X_norm = fold(X_norm, n_samples)
            return X_norm

    else:
        n_samples, n_chans, n_trials = theshapeof(X)
        weights = _check_weights(weights, X)
        if not weights.any():
            with np.errstate(divide="ignore"):
                N = ((np.sum(X ** 2, axis=0) / n_samples) ** -0.5)[np.newaxis]

            N[np.isinf(N)] = 0
            N[np.isnan(N)] = 0
            X_norm = X * N

        else:
            if weights.shape[0] != X.shape[0]:
                raise ValueError("Weight array should have same number of " +
                                 "columns as X")
            if weights.shape[1] == 1:
                weights = np.tile(weights, (1, n_chans))
            if weights.shape != X.shape:
                raise ValueError("Weight array should have be same shape as X")

            N = (np.sum(X ** 2 * weights, axis=0) /
                 np.sum(weights, axis=0)) ** -0.5
            N[np.isinf(N)] = 0
            N[np.isnan(N)] = 0
            X_norm = X * N

    if return_norm:
        return X_norm, np.sqrt(N)
    else:
        return X_norm


def matmul3d(X, mixin):
    """Apply mixing matrix to each trial of X.

    This is a simple wrapper or np.einsum.

    Parameters
    ----------
    X : array, shape=(n_samples, n_chans, n_trials)
    mixing : array, shape=(n_chans, n_components)

    Returns
    -------
    proj : array, shape=(n_samples, n_components, n_trials)
        Projection.

    """
    assert mixin.ndim == 2, "mixing matrix must be 2D"

    if X.ndim == 2:
        return X @ mixin
    elif X.ndim == 3:
        return np.einsum("sct,ck->skt", X, mixin)
    else:
        raise RuntimeError("X must be (n_samples, n_chans, n_trials)")


def _check_shifts(shifts, allow_floats=False):
    """Check shifts."""
    types = (int, np.int_)
    if allow_floats:
        types += (float, np.float64)
    if not isinstance(shifts, (np.ndarray, list, type(None)) + types):
        raise AttributeError("shifts should be a list, an array or an int")
    if isinstance(shifts, (list, ) + types):
        shifts = np.array(shifts).flatten()
    if shifts is None or len(shifts) == 0:
        shifts = np.array([0])

    n_shifts = np.size(shifts)

    return shifts, n_shifts


def _check_data(X):
    """Check data is numpy array and has the proper dimensions."""
    if not isinstance(X, (np.ndarray, list)):
        raise AttributeError("data should be a list or a numpy array")

    dtype = np.complex128 if np.any(np.iscomplex(X)) else np.float64
    X = np.asanyarray(X, dtype=dtype)
    if X.ndim > 3:
        raise ValueError("Data must be 3D at most")

    return X


def _check_weights(weights, X):
    """Check weights dimensions against X."""
    if not isinstance(weights, (np.ndarray, list)):
        if weights is not None:
            warnings.warn("weights should be a list or a numpy array.")
        weights = np.array([])

    weights = np.asanyarray(weights)
    if weights.size > 0:
        dtype = np.complex128 if np.any(np.iscomplex(weights)) else np.float64
        weights = np.asanyarray(weights, dtype=dtype)
        if weights.ndim > 3:
            raise ValueError("Weights must be 3D at most")
        if weights.shape[0] != X.shape[0]:
            raise ValueError("Weights should be the same n_times as X.")

        if X.ndim == 2 and weights.ndim == 1:
            weights = weights[:, np.newaxis]
        if X.ndim == 3:
            if weights.ndim == 2:
                weights = weights[:, np.newaxis, :]
            elif weights.ndim == 1:
                weights = weights[:, np.newaxis, np.newaxis]

            if weights.shape[-1] != X.shape[-1]:
                weights = np.tile(weights, (1, 1, X.shape[-1]))

        if weights.ndim > 1:
            if weights.shape[1] > 1 and weights.shape[1] != X.shape[1]:
                raise ValueError("Weights array should have a single column.")

        if np.any(np.abs(weights) > 1.):
            warnings.warn("weights should be between 0 and 1.")
            weights[np.abs(weights) > 1.] = 1.

    return weights


def _times_to_delays(lags, sfreq):
    """Convert a lags in seconds to delays."""
    if lags is None:
        return np.array([0])
    if not isinstance(sfreq, (int, float, np.int_)):
        raise ValueError("`sfreq` must be an integer or float")
    sfreq = float(sfreq)

    if not all([isinstance(ii, (int, float, np.int_)) for ii in lags]):
        raise ValueError("lags must be an integer or float")

    if len(lags) == 2 and sfreq != 1:
        tmin = lags[0]
        tmax = lags[1]

        if not tmin <= tmax:
            raise ValueError("tmin must be <= tmax")

        # Convert seconds to samples
        delays = np.arange(int(np.round(tmin * sfreq)),
                           int(np.round(tmax * sfreq) + 1), step=2)

    else:
        delays = lags

    return delays
